vec2 random2(vec2 st)
{
    vec2 s = vec2(dot(st, vec2(12.3456, 34.1415)),
                  dot(st, vec2(42.2154, 15.2854)));
    return fract(sin(s) * 45678.9) * 2.0 - 1.0;
}

float cellularNoise(vec2 st)
{
    ivec2 cell = ivec2(floor(st));
    vec2 f = st - vec2(cell);

    float minDist = 1.0;

    for (int y = -1; y <= 1; y++)
    {
        for (int x = -1; x <= 1; x++)
        {
            ivec2 localCell = cell + ivec2(x, y);
            vec2 localPoint = random2(vec2(localCell)) * 0.5 + 0.5 + vec2(x, y);
            minDist = min(minDist, length(localPoint - f));
        }
    }

    return minDist;
}

float fbmCellular(vec2 st)
{
    float val = 0.0;

    float p = 0.5;

    const float angle = 1.0;

    mat2 rotation = mat2(
        cos(angle), sin(angle),
       -sin(angle), cos(angle)
    );

    for (int i = 0; i < 5; i++)
    {
        val += cellularNoise(st) * p;
        p *= 0.5;
        st *= 2.0;

        st += vec2(1.181, 0.57);
        st = rotation * st;
    }

    return val;
}

void mainImage(out vec4 fragColor, in vec2 fragCoord)
{
    vec2 uv = fragCoord/iResolution.xy;
    // Střed souřadnic chceme uprostřed obrazu
    uv = uv * 2.0 - 1.0;
    // Korekce pro správný poměr stran
    uv.x *= iResolution.x / iResolution.y;

    vec3 col = vec3(fbmCellular(uv * 2.0));

    fragColor = vec4(col, 1.0);
}

