vec2 random2(vec2 st)
{
    vec2 s = vec2(dot(st, vec2(12.3456, 34.1415)),
                  dot(st, vec2(42.2154, 15.2854)));
    return fract(sin(s) * 45678.9) * 2.0 - 1.0;
}

float gradientNoise(vec2 st)
{
    vec2 cell = floor(st);
    vec2 f = fract(st);

    vec2 s00 = random2(cell);
    vec2 s01 = random2(cell + vec2(0, 1));
    vec2 s10 = random2(cell + vec2(1, 0));
    vec2 s11 = random2(cell + vec2(1, 1));

    float d00 = dot(s00, f);
    float d01 = dot(s01, f - vec2(0, 1));
    float d10 = dot(s10, f - vec2(1, 0));
    float d11 = dot(s11, f - vec2(1, 1));

    vec2 u = smoothstep(0.0, 1.0, f);

    float noise = mix(mix(d00, d10, u.x), mix(d01, d11, u.x), u.y);
    return noise * 0.5 + 0.5;
}

// "Barevný" gradient noise (tři hodnoty v -1..1)
vec3 gradientNoise3(vec2 st)
{
    // Souřadnice šumů pro různé složky jsou různé,
    // aby byl výsledný šum barevný
    vec3 result = vec3(
        gradientNoise(st),
        gradientNoise(st + vec2(1.11, 2.22)),
        gradientNoise(st + vec2(3.33, 4.44))
    );
    // Hodnoty převedeme do rozsahu -1..1
    result = result * 2.0 - 1.0;
    // Hodnoty také něčím vynásobíme, aby byly více kontrastní
    result = clamp(result * 1.7, -1.0, 1.0);
    return result;
}

// "Barevný" fbm (tři hodnoty v -1..1)
vec3 fbm3(vec2 st)
{
    vec3 val = vec3(0.0);

    float p = 0.5;

    const float angle = 1.0;

    mat2 rotation = mat2(
        cos(angle), sin(angle),
       -sin(angle), cos(angle)
    );

    for (int i = 0; i < 5; i++)
    {
        val += gradientNoise3(st) * p;
        p *= 0.5;
        st *= 2.0;

        st += vec2(1.181, 0.57);
        st = rotation * st;
    }

    return val;
}

vec3 cellularNoiseColored(vec2 st)
{
    ivec2 cell = ivec2(floor(st));
    vec2 f = st - vec2(cell);

    // Pamatujeme si, z které buňky pochází nejbližší bod
    ivec2 closestCell = ivec2(0, 0);
    float minDist = 1.0;

    for (int y = -1; y <= 1; y++)
    {
        for (int x = -1; x <= 1; x++)
        {
            ivec2 localCell = cell + ivec2(x, y);
            // Vstup do fbm spočítáme z buňky plus posunutí podle času
            vec2 fbmCoords = vec2(localCell) + iTime * vec2(0.2, 0.1);
            // Souřadnice bodu získáme z fbm3 (vracející v rozsahu -1..1),
            // které něčím pronásobíme, aby byly souřadnice zahímavější,
            // ale také clampneme dovnitř buňky (-0.99..0.99)
            // a nakonec převedeme do 0..1 a přičteme souřadnice rohu buňky
            vec2 localPoint = clamp(fbm3(fbmCoords).xy * 2.0, -0.99, 0.99)
                * 0.5 + 0.5 + vec2(x, y);
            float dist = length(localPoint - f);
            if(dist < minDist)
            {
                minDist = dist;
                closestCell = localCell;
            }
        }
    }
    
    // Zabarvíme podle buňky nebližšího bodu...
    vec3 color = gradientNoise3(vec2(closestCell) * 1.5) * 0.5 + 0.5;
    // ...a také podle jeho vzdálenosti
    color *= clamp(sqrt(minDist), 0.0, 1.0);
    return color;
}

void mainImage(out vec4 fragColor, in vec2 fragCoord)
{
    vec2 uv = fragCoord/iResolution.xy;
    // Střed souřadnic chceme uprostřed obrazu
    uv = uv * 2.0 - 1.0;
    // Korekce pro správný poměr stran
    uv.x *= iResolution.x / iResolution.y;

    vec3 col = cellularNoiseColored(uv * 2.0);

    fragColor = vec4(col, 1.0);
}

