// Vše hlavní je ve funkci mainImage, na řádku 158

vec2 random2(vec2 st)
{
    vec2 s = vec2(
        dot(st, vec2(12.3456, 34.1415)),
        dot(st, vec2(42.2154, 15.2854))
    );
    return fract(sin(s) * 45678.9) * 2.0 - 1.0;
}

float gradientNoise(vec2 st)
{
    vec2 cell = floor(st);
    vec2 f = fract(st);

    vec2 s00 = random2(cell);
    vec2 s01 = random2(cell + vec2(0, 1));
    vec2 s10 = random2(cell + vec2(1, 0));
    vec2 s11 = random2(cell + vec2(1, 1));

    float d00 = dot(s00, f);
    float d01 = dot(s01, f - vec2(0, 1));
    float d10 = dot(s10, f - vec2(1, 0));
    float d11 = dot(s11, f - vec2(1, 1));

    vec2 u = smoothstep(0.0, 1.0, f);

    float noise = mix(
        mix(d00, d10, u.x),
        mix(d01, d11, u.x),
        u.y
    );
    return noise * 0.5 + 0.5;
}

float fbm(vec2 st)
{
    float val = 0.0;

    float p = 0.5;

    const float angle = 1.0;

    mat2 rotation = mat2(
        cos(angle), sin(angle),
       -sin(angle), cos(angle)
    );

    for (int i = 0; i < 5; i++)
    {
        val += gradientNoise(st) * p;
        p *= 0.5;
        st *= 2.0;

        st += vec2(1.181, 0.57);
        st = rotation * st;
    }

    return val;
}

vec4 brick(vec2 uv)
{
    // Chceme vidět v rozsahu -1..1
    // více řad cihel, navíc cheme,
    // aby byly cihly obdélníkové
    vec2 brickuv = uv * vec2(2.0, 4.0);

    int row = int(floor(brickuv.y));

    // Každou druhou řadu cihel posuneme
    if((row & 1) == 0)
        brickuv.x += 0.5;

    vec2 f = fract(brickuv);

    // Převedeme je do rozsahu -1..1
    f = f * 2.0 - 1.0;

    // Souřadnice, ze kterých bereme šum
    // pro cihly, jsou různé pro každou buňku
    vec2 noiseuv = brickuv * 2.0
        + floor(brickuv) * 42.2;
    // Náhodný šum převedený do rozsahu -1..1
    vec2 noise = vec2(
        fbm(noiseuv),
        fbm(noiseuv + vec2(12.244, 321.25))
    ) * 2.0 - 1.0;

    f = abs(f + noise * 0.25);

    float height = max(f.x, f.y);
    height = max((height - 0.75) * 4.0, 0.0);
    height = pow(height, 3.0) * 3.0;
    height = 1.0 - clamp(height, 0.0, 1.0);

    // Šum pro maltu v rozsahu 0..1
    // Používáme původní uv, nikoliv brickuv
    float backgroundNoise = fbm(uv * 64.0);

    height = clamp(
        height + backgroundNoise * 0.25,
        0.0, 1.0);

    // Šum pro cihly v rozsahu 0..1
    float foregroundNoise = fbm(uv * 16.0);

    // Šum umocníme na druhou
    foregroundNoise *= foregroundNoise;

    height = clamp(height
        - height * (foregroundNoise)* 0.5,
        0.0, 1.0);

    // Výsledek obarvíme
    vec3 color = vec3(0.4);

    vec3 brickColor = vec3(0.8, 0.55, 0.5)
        + vec3(random2(floor(brickuv)), 0.0)
        * vec3(0.15, 0.075, 0.0);

    color = mix(color, brickColor, height);

    return vec4(color, height);
}

float getHeight(vec2 uv) {
    // Výšku pronásobíme deseti, aby byly normály zajímavější
    return brick(uv).w * 10.0;
}

vec3 computeNormal(vec2 uv)
{
    const float diff = 0.001;
    float here = getHeight(uv);

    vec3 n = vec3(
        here - getHeight(uv + vec2(diff, 0.0)),
        here - getHeight(uv + vec2(0.0, diff)),
        1.0
    );
    return normalize(n);
}

// Transformuje souřadnice na obrazovce v rozsahu 0..1
// na souřadnice v prostoru cihel
vec2 screenSpaceToWorldSpace(vec2 uv)
{
    // Převedeme uv do rozsahu -1..1
    uv = uv * 2.0 - 1.0;
    // "Natáhneme" X tak, aby byl zachován poměr stran
    uv.x *= iResolution.x / iResolution.y;
    // Nyní je tedy čtverec v souřadnicích i čtvercem na monitoru
    return uv;
}

void mainImage(out vec4 fragColor, in vec2 fragCoord)
{
    vec2 uv = screenSpaceToWorldSpace(fragCoord/iResolution.xy);

    // Spočítáme aktuální barvu a výšku povrchu (výška je surface.w)
    vec4 surface = brick(uv);
    
    // Toto ignorujte. Proč to tu ale je? Viz komentář na konci shaderu.
    //#define GAMMA_CORRECT
    
    // Toto ignorujte.
    #ifdef GAMMA_CORRECT
    surface.rgb = pow(surface.rgb, vec3(2.2));
    #endif

    // Spočítáme normálu v tomto bodě
    vec3 normal = computeNormal(uv);
    
    // Zde je uložena pozice pixelu v prostoru.
    // Komponenta Z je nastavena na výšku povrchu v tomto bodě.
    // V tomto případě by šla by použít i nula a shader by se choval
    // zhruba stejně. To často také dělají hry, když používají normalmapy.
    // Co se pozice týče, počítají stále s placatým povrchem, a mění jen normálu.
    vec3 position = vec3(uv.x, uv.y, surface.w * 0.1);
    // Difůzní barva povrchu
    vec3 diffuseColor = surface.rgb;
    // Spekulární barva - všude stejná
    vec3 specularColor = vec3(0.5);
    // Lesklost povrchu
    float phong_a = 32.0;
    
    // Pozice světla. Všimněte si, že na souřadnice myši aplikujeme stejnou korekci
    // jako na souřadnice na obrazovce. Poslední komponenta je výška světla nad povrchem.
    vec3 lightPosition = vec3(screenSpaceToWorldSpace(iMouse.xy/iResolution.xy), 1.2);
    
    // Intenzita či barva světla
    vec3 lightIntensity = vec3(1.0, 0.75, 0.45);
    
    // Pozice kamery ve světě
    vec3 cameraPosition = vec3(0.0, 0.0, 3.0);
    
    // Normalizovaný view vector, tedy vektor z místa odpovídajícímu tomuto pixelu směrem k pozorovateli
    vec3 view = normalize(cameraPosition - position);
    
    // Normalizovaný light vector, tedy z pixelu směrem ke světlu
    vec3 lightVec = normalize(lightPosition - position);
    // Reflection vector, tedy kam by se světlo zrcadlově odrazilo z tohoto místa
    vec3 reflectionVec = reflect(-lightVec, normal);
    
    // Vzdálenost od pixelu ke světlu
    float distanceToLight = length(lightPosition - position);
    
    // O kolik se intenzita světla zmenší vlivem vzdálenosti
    float attenuatedIntensity = 1.0 / (distanceToLight * distanceToLight);
    
    // Spočítáme difůzní světlo
    // Všimněte si, že skalární součin (dot) je obalen do max(..., 0.0),
    // aby nebyl nikdy záporný.
    vec3 lightDiffuse =
        lightIntensity * attenuatedIntensity
        * diffuseColor
        * max(dot(lightVec, normal), 0.0);
    
    // Spočítáme spekulární světlo
    vec3 lightSpecular =
        lightIntensity * attenuatedIntensity
        * specularColor
        * max(dot(lightVec, normal), 0.0)
        * pow(max(dot(view, reflectionVec), 0.0), phong_a);
    
    // Finální světlo co doputuje k pozorovateli je součet difůzního a spekulárního světla
    vec3 lightFinal = lightDiffuse + lightSpecular;
    
    // Zobrazíme
    fragColor = vec4(lightFinal, 1.0);
    
    // Četli jste si už čtvrtý díl seriálu a zdá se vám,
    // že by tady měla být gamma korekce? Máte pravdu!
    // Ve třetím jsem ji vynechal kvůli stručnosti, ale
    // jelikož zde už počítáme konkrétní hodnoty světla,
    // měli bychom ji použít.
    //
    // !ALE! stejně tak bychom měli aplikovat opačnou korekci
    // na spočítanou barvu cihel!
    // Tu jsme totiž odladili tak, aby vypadala dobře, když
    // se zobrazuje na motinor jak leží běží.
    // Pokud ji chceme převést na intenzitu odraženého světla,
    // musíme ji umocnit na hodnotu gammy (zde 2.2),
    // viz řádek 170.
    //
    // Pokud chcete zapnout správnou gamma korekci,
    // odkomentujte řádek 166.
    //
    // A jestli přicházíte z daleké budoucnosti, tedy
    // pátého dílu seriálu, a chybí vám v počítání difůzního
    // světla vydělení pi, také máte pravdu.
    
    #ifdef GAMMA_CORRECT
    fragColor = vec4(pow(lightFinal, vec3(1.0/2.2)), 1.0);
    #endif
}

