///////////////////////////////////////////////////////////////////
//
// Zde je kód pro progresivní vykreslování a nějaké pomocné funkce.
// Doporučuji si aspoň jednou pořádně projít, co všechno zde je.
// Hlavně vás zajímá funkce "computeImage" na řádku 207.
//
///////////////////////////////////////////////////////////////////

// Pokud chcete, aby se při kliknutí myší
// začalo vykreslovat od začátku,
// odkomentujte následující řádek:
//#define RESET_ON_MOUSE_MOVE

#define BOUNCES 2
#define SUN
#define AA
//#define GLOW

// Bude se vám hodit pi
#define PI 3.1415926535897932384626433832795

// Tuto vzdálenost budeme považovat za nekonečno
#define FAR 50.0

// Náhodné funkce vracející hodnoty v rozsahu -1..1
vec2 random2(vec2 st)
{
    vec2 s = vec2(dot(st, vec2(12.3456, 34.1415)),
                  dot(st, vec2(42.2154, 15.2854)));
    return fract(sin(s) * 45678.9) * 2.0 - 1.0;
}

vec3 random3(vec3 st)
{
    vec3 s = vec3(dot(st, vec3(12.3456, 34.1415, 21.4789)),
                  dot(st, vec3(42.2154, 15.2854, 33.5478)),
                  dot(st, vec3(32.2487, 27.7891, 14.4751)));
    return fract(sin(s) * 45678.9) * 2.0 - 1.0;
}

// Generování náhodného bodu v jednotkové kouli pomocí rejection samplingu
vec3 randomPointInUnitSphere(vec3 seed)
{
    const int tries = 10;
    for(int i = 0; i < tries; i++)
    {
        vec3 dir = random3(seed + vec3(i));
        if(length(dir) <= 1.0)
            return dir;
    }
    return normalize(random3(seed + vec3(tries)));
}

// Generování náhodného bodu na jednotkové polokouli, která svou
// plochou stranou leží na povrchu s danou normálou
// Toto se vám bude hodit ;)
vec3 randomPointOnUnitHemisphere(vec3 seed, vec3 normal)
{
    vec3 d = randomPointInUnitSphere(seed);
    if(dot(d, normal) > 0.0)
        return d;
    return -d;
}

//
//
// Definice scény a kamery
//
//

float sdfSphere(vec3 center, float radius, vec3 p)
{
    return length(center - p) - radius;
}

// Zdroj: https://iquilezles.org/www/articles/distfunctions/distfunctions.htm
float sdfBox(vec3 center, vec3 size, vec3 p)
{
    vec3 q = abs((p - center)) - size;
    return length(max(q,0.0)) + min(max(q.x,max(q.y,q.z)),0.0);
}

float sceneSdf(vec3 p)
{
    float d = p.y;
    // Boční zdi
    d = min(d, sdfBox(vec3(-3.0, 1.0, 0.0), vec3(0.25, 1.25, 3.0), p));
    d = min(d, sdfBox(vec3(3.0, 1.0, 0.0), vec3(0.25, 1.25, 3.0), p));
    // Strop
    d = min(d, sdfBox(vec3(0.0, 2.5, 0.0), vec3(3.25, 0.25, 3.0), p));
    // Zadní zeď
    d = min(d, sdfBox(vec3(0.0, 1.2, -2.0), vec3(3.0, 1.5, 3.0), p));
    // Koule
    d = min(d, sdfSphere(vec3(0.0, 0.8, 2.0), 0.8, p));
    
    #ifdef GLOW
    // Koule 2
    d = min(d, sdfSphere(vec3(1.8, 1.3, 2.0), 0.4, p));
    #endif
    
    return d;
}

vec3 getSceneDiffuseColor(vec3 p)
{
    vec3 color = vec3(0.75);
    
    // Zelená podlaha
    if(p.y < 0.03)
    {
    	color = vec3(0.3, 1.0, 0.3);
    }
    
    // Červená zeď
    if(sdfBox(vec3(-3.0, 1.0, 0.0), vec3(0.25, 1.25, 3.0), p) < 0.02)
    {
        color = vec3(1.0, 0.1, 0.01);
    }
    
    return color;
}

vec3 getSky(vec3 direction)
{
    #ifdef GLOW
	return mix(vec3(0.8, 0.9, 1.0), vec3(0.5, 0.7, 1.0), max(direction.y, 0.0)) * 0.1;
    #else
    return mix(vec3(0.8, 0.9, 1.0), vec3(0.5, 0.7, 1.0), max(direction.y, 0.0));
    #endif
}

vec3 sunlightDirection = normalize(vec3(1.9, 0.8, 1.0));
#ifdef GLOW
vec3 sunlightColor = vec3(0.0);
#else
vec3 sunlightColor = vec3(1.0, 0.9, 0.8) * 4.0;
#endif

vec3 cameraPosition = vec3(0.0, 1.0, 7.3);
vec3 cameraLookAt = vec3(0.0, 1.0, 0.0);
float cameraFov = 80.0;

//
//
// Pomocné funkce pro vykreslování
//
//

// Raymarchovací funkce
// Vstupy:
// origin - počátek paprsku
// direction - směr paprsku (vektor normalizovaný na velikost 1)
// Výstup: vzdálenost, kterou musí paprsek urazit, než narazí na kolizi
// nebo hodnota větší nebo rovnou než FAR, pokud do vzdálenosti FAR na
// žádnou kolizi nenarazil.
// Nezapomeňte smyčku ukončit, pokud hodnota SDF klesne pod nějaké epsilon
// nebo pokud se provede příliš mnoho iterací.
float raymarch(vec3 origin, vec3 direction)
{
    const float epsilon = 0.003;
    float marched = 0.0;
    int i;
    const int iterations = 256;
    for(i = 0; i < iterations; i++)
    {
        float dist = sceneSdf(origin + direction * marched);
        if(dist < epsilon)
            break;
        marched += dist;
        if(marched > FAR)
            break;
    }
    if(i >= iterations)
        return (FAR + 1.0);
    return marched;
}

vec3 computeNormal(vec3 p)
{
    const float epsilon = 0.001;
    float here = sceneSdf(p);
    return normalize(vec3(
        sceneSdf(p + epsilon * vec3(1.0, 0.0, 0.0)) - here,
        sceneSdf(p + epsilon * vec3(0.0, 1.0, 0.0)) - here,
        sceneSdf(p + epsilon * vec3(0.0, 0.0, 1.0)) - here
    ));
}

// forward: normalizovaný vektor dopředu
// uv: souřadnice pixelu v -1..1
// hfov: horizontální zorný úhel ve stupních
vec3 getRayDirection(vec3 forward, vec2 uv, float hfov)
{
    // Z "forward" vygenerujeme i vektory vpravo a nahoru, vše normalizované a navzájem kolmé
    vec3 right = normalize(cross(forward, vec3(0.0, 1.0, 0.0)));
    vec3 up = normalize(cross(right, forward));
    // Převedeme hfov na radiány
    hfov = hfov / 180.0 * PI;
    // Spočítáme vzdálenost roviny od kamery
    float dist = 1.0 / tan(hfov * 0.5);
    // Přeškálujeme "up" tak, aby byl zachován poměr stran
    up *= iResolution.y / iResolution.x;
    vec3 ray = forward * dist + right * uv.x + up * uv.y;
    return normalize(ray);
}

void computeImage(out vec4 fragColor, in vec2 fragCoord)
{
    // Souřadnice pixelu v rozsahu -1..1
    #ifdef AA
    vec2 uv = (fragCoord + random2(vec2(float(iFrame), 0.0)) * 0.5) / iResolution.xy * 2.0 - 1.0;
    #else
    vec2 uv = fragCoord / iResolution.xy * 2.0 - 1.0;
    #endif
    
    vec3 rayDirection = getRayDirection(
        normalize(cameraLookAt - cameraPosition),
        uv, cameraFov);

    float depth = raymarch(cameraPosition, rayDirection);
    
    // Souřadnice dopadu paprsku
    vec3 hit = cameraPosition + rayDirection * depth;
    
    vec3 lightFactor = vec3(1.0);
    vec3 lightOut = vec3(0.0);
    
    for(int i = 0; i <= BOUNCES; i++)
    {
        if(depth >= FAR)
        {
            lightOut += lightFactor * getSky(rayDirection);
            break;
        }
        
        // Normála povrchu v bodě dopadu
        vec3 normal = computeNormal(hit);
        vec3 diffuseColor = getSceneDiffuseColor(hit);
        vec3 origin = hit + normal * 0.02;
        
        #ifdef SUN
        // Příme osvětlení od slunce
        float shadowDepth = raymarch(origin, sunlightDirection);
        if(shadowDepth >= FAR)
        {
            vec3 sun = diffuseColor / PI * max(dot(normal, sunlightDirection), 0.0) * sunlightColor;
            lightOut += sun * lightFactor;
        }
        #endif
        
        #ifdef GLOW
        if(sdfSphere(vec3(0.0, 0.8, 2.0), 0.8, hit) < 0.02)
        {
            lightOut += lightFactor * vec3(1.0, 0.8, 0.01) * 4.0;
        }
        #endif
        
        if(i == BOUNCES)
            break;
        
        // Pokud jsme nepřekročili maximální povolený počet odrazů,
        // pokračujeme v počítání nepřímého světla
        rayDirection = randomPointOnUnitHemisphere(hit + vec3(float(iFrame), 0.0, 0.0), normal);
        depth = raymarch(origin, rayDirection);
        hit = origin + rayDirection * depth;
        lightFactor *= diffuseColor / PI * max(dot(normal, rayDirection), 0.0);
        lightFactor *= 2.0 * PI;
    }
    
    fragColor.rgb = lightOut;
}

void mainImage(out vec4 fragColor, in vec2 fragCoord)
{
    // Souřadnice pixelu v rozsahu 0..1
    vec2 uv = fragCoord/iResolution.xy;
    
    // V pixelu na těchto souřadnicích si budeme
    // pamatovat pomocné hodnoty
    vec2 memory_texcoord = vec2(0.5) / iResolution.xy;
    
    // Načteme zapamatované hodnoty
    vec4 memory = texture(iChannel0, memory_texcoord);
    
    // Pokud je aktuální pixel na pozici, ve kterém jsou v textuře
    // pomocné hodnoty, místo počítání obrazu uložíme pomocné hodnoty
    if(uv == memory_texcoord)
    {
        fragColor = vec4(iMouse.x, iMouse.y, iResolution.xy);
        return;
    }

    // Načteme starou zprůměrovanou hodnotu
    // V kanálu W si ukládáme počet zprůměrovaných snímků,
    // ze kterých je hodnota tvořena
    vec4 oldColor = texture(iChannel0, uv);
    
    // Chceme začít vykreslovat od začátku pokud je splněno aspoň jedno:
    // - změnilo se rozlišení obrazu
    // - pohla se myš
    // - iTime je blízko nuly - účel je, že pokud zapausujete překreslování a resetujete tlačítkem
    //   čas na nulu, tak po překompilování shaderu uvidíte nový výsledek.
    //   Jinak by se do zprůměrovaného obrazu starého shaderu jen začaly pomalu průměrovat výsledky
    //   nově zkompilovaného.
    if(iResolution.xy != memory.zw
    #ifdef RESET_ON_MOUSE_MOVE
     || iMouse.xy != memory.xy
    #endif
     || iTime <= 0.01)
    {
        // Když chceme začít vykreslovat od začátku, vynulujeme oldColor
        oldColor = vec4(0.0);
    }
    
    // Spočítáme nový obraz
    vec4 color;
    computeImage(color, fragCoord);
    
    // Nový obraz připočítáme k zprůměrovanému starému
    // (zprůměrovaný nový obraz) =
    // (zprůměrovaný starý obraz) * (počet snímků starého obrazu) + (nový obraz)
    // To celé vydělené celkovým počtem snímků, které jsou do hodnoty započítány,
    // tedy počtem snímků ve staré hodnotě + 1
    color.rgb = (oldColor.rgb * oldColor.w + color.rgb) / (oldColor.w + 1.0);
    
    // Výsledek uložime
    fragColor = vec4(color.rgb, oldColor.w + 1.0);
}

